
/*
 *  SIMD Smith-Waterman function for x86 platform
 *
 *  Technology: Intel MMX
 *  Algorithm:  Diagonal
 *  Precision:  Word
 *
 *  Copyright (C) 2003 Arpith Chacko Jacob
 *  Email: jarpith@acm.org
 *  Beta release 1.0.0.2
 *
 *  Start date:  26 june 2003
 *  Last Update: 08 september 2003
 *
 */

#define QUERY		8(%ebp)
#define M			12(%ebp)
#define SEQ			16(%ebp)
#define N			20(%ebp)
#define Q_PROFILE	24(%ebp)
#define BIAS		28(%ebp)
#define GAP_INIT	32(%ebp)
#define GAP_EXT		36(%ebp)
#define FIH			40(%ebp)

#define SIMD_BIAS	4 * SIMD_SIZE(%esp)
#define SIMD_QPP_MEM	0(%esp)

#define SIMD_SIZE	4

.macro PMAXUW		SOURCE, DEST
	psubusw			\SOURCE, \DEST
	paddusw			\SOURCE, \DEST
.endm

.global _sw_simd_mmx_diag_w, sw_simd_mmx_diag_w


.text

/* 
 * function _sw_simd_mmx_diag_w
 * C prototype
 * int sw_simd_mmx_diag_w(char *query, int m, char *seq, int n, short *query_profile,
 *                        int bias, int gap_init, int gap_ext, short *fih);
 *
 */

sw_simd_mmx_diag_w:
_sw_simd_mmx_diag_w:

	/* Function preprocessing */
	enter		$0, $0
	pusha

	/* Calculate the adjusted (with dummy chars) length of the query sequence */
	add			$(SIMD_SIZE * 2 - 2), M

	/* Align stack pointer for SIMD_SIZE (2^x) operations */
	movl		%esp, %eax
	sub			$4, %eax			    /* Size of push eax */
	and			$(SIMD_SIZE * 2 - 1), %eax	/* Calculate alignment offset */
	sub			%eax, %esp		    	/* Add offset */
	push		%eax			        /* Store offset */

	/* Load bias */
	movw		BIAS, %ax
	shl			$16, %eax
	movw		BIAS, %ax
	push		%eax                 	/* Store BIAS in SIMD_SIZE aligned */
	push		%eax					/* location 					   */

	/* Space for SIMD_SM_MEM */
	sub			$(SIMD_SIZE * 4), %esp

	/* Load gap_init, gap_ext */
	pxor		%mm6, %mm6
	pxor		%mm7, %mm7

	movw		GAP_INIT, %ax
	movw		%ax, 2 + GAP_INIT
	movd		GAP_INIT, %mm6
	punpckldq	%mm6, %mm6

	movw		GAP_EXT, %ax
	movw		%ax, 2 + GAP_EXT
	movd		GAP_EXT, %mm7
	punpckldq	%mm7, %mm7

	/* for (i = 0; i < m + 2 * SIMD_SIZE - 2; i++) {
		 F(i) = 0;
	 	 H(i) = 0;
	   }
	*/

	xor			%eax, %eax
	movl		M, %ebx

	movl		FIH, %ecx

init_loop:
	movl		$0, (%ecx, %eax, 4)

	inc			%eax
	cmp			%ebx, %eax
	jl			init_loop

	/* score = 0 */
	pxor		%mm0, %mm0

	/* for (i = 0; i < n/4; i++) */
	xor			%eax, %eax
	movl		N, %ebx
	shr			$2, %ebx			    /* SIMD = 8, shift right thrice */

	movl		SEQ, %edi

outer_loop:

	/* Save outer loop counters */
	push		%eax
	push		%ebx

	/* Clear MMX registers */
	pxor		%mm1, %mm1
	pxor		%mm4, %mm4
	pxor		%mm5, %mm5

	/* fk <- [ F(k - 1), 0, 0, 0 ] */
	movw		((SIMD_SIZE - 1) * 4)(%ecx), %ax
	movd		%eax, %mm2
	psllq		$48, %mm2

	/* hk <- [ H(k - 1), 0, 0, 0 ] */
	movw		((SIMD_SIZE - 1) * 4 + 2)(%ecx), %ax
	movd		%eax, %mm3
	psllq		$48, %mm3

	/* Calculate query_profile scores for next k set */
	movl		Q_PROFILE, %eax

	movzx		0(%edi), %edx
	imul		M, %edx
	shl			$1, %edx
	add			$((SIMD_SIZE - 1 - 0) * 2), %edx
	add			%eax, %edx
	movl		%edx, 8 + SIMD_QPP_MEM

	movzx		1(%edi), %edx
	imul		M, %edx
	shl			$1, %edx
	add			$((SIMD_SIZE - 1 - 1) * 2), %edx
	add			%eax, %edx
	movl		%edx, 4 + 8 + SIMD_QPP_MEM

	movzx		2(%edi), %edx
	imul		M, %edx
	shl			$1, %edx
	add			$((SIMD_SIZE - 1 - 2) * 2), %edx
	add			%eax, %edx
	movl		%edx, 8 + 8 + SIMD_QPP_MEM

	movzx		3(%edi), %edx
	imul		M, %edx
	shl			$1, %edx
//	add			$((SIMD_SIZE - 1 - 3) * 2), %edx
//						^^^ ---> evaluates to zero
	add			%eax, %edx
	movl		%edx, 12 + 8 + SIMD_QPP_MEM

    xor			%esi, %esi

	/* for (j = 0; j < m+k-1; j++) */
	xor			%eax, %eax
	movl		M, %ebx
	sub			$(SIMD_SIZE - 1), %ebx

	/* mm0 - scorek
	   mm1 - ek
	   mm2 - fk
	   mm3 - hk
	   mm4 - hck
	   mm5 - hdk
	   mm6 - gap_init
	   mm7 - gap_ext
	 */

inner_loop:
	/* Push loop counters */
	push		%ebx
	push		%eax

	/* ek <- max { ek, hck - qk } - rk */
	psubusw		%mm6, %mm4
	PMAXUW		%mm4, %mm1
	psubusw		%mm7, %mm1

	/* fk <- max { fk, hk - qk } - rk */
	movq		%mm3, %mm4
	psubusw		%mm6, %mm4
	PMAXUW		%mm4, %mm2
	psubusw		%mm7, %mm2

	/* hck <- max { 0, ek, hdk + sim_mat(), fk }
	   Sim matrix in mm4
	   s(k * i + 0, j + k - 0 )
	   s(k * i + 1, j + k - 1 )
	   ..
	   ..
	   s(k * i + 7, j + k - 7 )
	 */

	movl		16 + SIMD_QPP_MEM, %edx
	movzx		(%edx, %esi, 2), %ax

	shl			$16, %eax

	movl		4 + 16 + SIMD_QPP_MEM, %edx
	movzx		(%edx, %esi, 2), %ax

	movl		%eax, BIAS				/* Reuse stack var as temp var */

	movl		8 + 16 + SIMD_QPP_MEM, %edx
	movzx		(%edx, %esi, 2), %ax

	shl			$16, %eax

	movl		12 + 16 + SIMD_QPP_MEM, %edx
	movzx		(%edx, %esi, 2), %ax

	movd		%eax, %mm4
	punpckldq	BIAS, %mm4

	/* Increment query counter to next k set */
	inc			%esi

	/* hck <- max { 0, ek, hdk + sim_mat(), fk }
	   hdk + sim_mat()
	 */
	paddusw		%mm5, %mm4

	/* hdk + sim_mat() - BIAS */
	psubusw		16 + SIMD_BIAS, %mm4

	/* hck <- max { hck, ek, fk } */
	PMAXUW		%mm1, %mm4
	PMAXUW		%mm2, %mm4

	/* scorek <- max { scorek, hck } */
	PMAXUW		%mm4, %mm0

	/* hdk <- hk */
	movq		%mm3, %mm5

	/* Pop loop counter for use below */
	pop			%eax

	/* The following two code segments considerably slows down this loop,
	 * when compared to the SSE/SSE2 implementations due to the 64-bit memory
	 * references. However, we have no alternative as the MMX instruction set
	 * is severely limited.
	 */
	/* F(j) <- fk(8)
	   fk <- F(j + k) | fk(1..7)
	 */
	movd		%mm2, %edx
	psrlq		$16, %mm2
	movw		%dx, (%ecx, %eax, 4)
	movq		%mm2, GAP_INIT
	movw		(SIMD_SIZE * 4)(%ecx, %eax, 4), %dx
	movw		%dx, 6 + GAP_INIT
	movq		GAP_INIT, %mm2

	/* H(j) <- hck(8)
	   hk <- H(j + k) | hck(1..7)
	 */
	movq		%mm4, %mm3
	movd		%mm3, %ebx
	psrlq		$16, %mm3
	movw		%bx, 2(%ecx, %eax, 4)
	movq		%mm3, GAP_INIT
	movw		(SIMD_SIZE * 4 + 2)(%ecx, %eax, 4), %bx
	movw		%bx, 6 + GAP_INIT
	movq		GAP_INIT, %mm3

	/* Pop loop limit */
	pop			%ebx

	inc			%eax
	cmp			%ebx, %eax
	jl			inner_loop


	/* Restore outer loop counters */
	pop			%ebx
	pop			%eax

	/* Increment seq counter to next k set */
	add			$(SIMD_SIZE), %edi

	inc			%eax
	cmp			%ebx, %eax
	jl			outer_loop

	/* Calculate final score */
	movq		%mm0, %mm1
	psrlq		$32, %mm0
	PMAXUW		%mm1, %mm0

	movq		%mm0, %mm1
	psrlq		$16, %mm0
	PMAXUW		%mm1, %mm0

	xor			%eax, %eax
	movd		%mm0, %ebx
	mov			%bx, %ax
	movl		%eax, BIAS				/* Reuse stack variable */

#ifdef	DEBUG
	/* Done using MMX */
	emms
#endif

    /* Remove SIMD_BIAS/SIMD_SM_MEM */
	add			$((SIMD_SIZE + 2) * 4), %esp

	pop			%eax			        /* Get alignment offset */
	add			%eax, %esp		    	/* Add offset */

	/* Function post processing */
	popa
	mov			BIAS, %eax
	leave
	ret
