
/*
 *  SIMD Smith-Waterman function for x86 platform
 *
 *  Technology: Intel MMX
 *  Algorithm:  Horizontal
 *  Precision:  Byte
 *
 *  Copyright (C) 2003 Arpith Chacko Jacob
 *  Email: jarpith@acm.org
 *  Beta release 1.0.0.2
 *
 *  Start date:  28 june 2003
 *  Last Update: 08 september 2003
 *
 */

#define QUERY		8(%ebp)
#define M			12(%ebp)
#define SEQ			16(%ebp)
#define N			20(%ebp)
#define Q_PROFILE	24(%ebp)
#define BIAS		28(%ebp)
#define GAP_INIT	32(%ebp)
#define GAP_EXT		36(%ebp)
#define FIH			40(%ebp)

#define SIMD_BIAS	0(%esp)

#define SIMD_SIZE	8

.global _sw_simd_mmx_horiz_b, sw_simd_mmx_horiz_b

.macro PMAXUB		SOURCE, DEST
	psubusb			\SOURCE, \DEST
	paddusb			\SOURCE, \DEST
.endm

.text

/* 
 * function _sw_simd_mmx_horiz_b
 * C prototype
 * int sw_simd_mmx_horiz_b(char *query, int m, char *seq, int n, char *query_profile,
 *                         int bias, int gap_init, int gap_ext, char *fih);
 *
 */

sw_simd_mmx_horiz_b:
_sw_simd_mmx_horiz_b:

	/* Function preprocessing */
	enter		$0, $0
	pusha

	/* Align stack pointer for SIMD_SIZE (2^x) operations */
	movl		%esp, %eax
	sub			$4, %eax			    /* Size of push eax */
	and			$(SIMD_SIZE - 1), %eax	/* Calculate alignment offset */
	sub			%eax, %esp		    	/* Subtract offset */
	push		%eax			        /* Store offset */

	/* Load bias */
	movb		BIAS, %al
	movb		%al, %ah
	shl			$16, %eax
	movb		BIAS, %al
	movb		%al, %ah
	push		%eax                 	/* Store BIAS in SIMD_SIZE aligned */
	push		%eax					/* location 					   */

	/* Load gap_init, gap_ext */
	pxor		%mm6, %mm6
	pxor		%mm7, %mm7

	movb		GAP_INIT, %al
	movb		%al, %ah
	movw		%ax, GAP_INIT
	movw		%ax, 2 + GAP_INIT
	movd		GAP_INIT, %mm6
	punpckldq	%mm6, %mm6

	movb		GAP_EXT, %al
	movb		%al, %ah
	movw		%ax, GAP_EXT
	movw		%ax, 2 + GAP_EXT
	movd		GAP_EXT, %mm7
	punpckldq	%mm7, %mm7

	/* for (i = 0; i < m; i++) {
		 F(i) = 0;
	 	 H(i) = 0;
	   }
	*/

	xor			%eax, %eax
	movl		M, %ebx

	movl		FIH, %ecx

init_loop:
	movw		$0, (%ecx, %eax, 2)

	inc			%eax
	cmp			%ebx, %eax
	jl			init_loop

	/* score = 0 */
	pxor		%mm0, %mm0

	/* for (i = 0; i < n; i++) */
	xor			%eax, %eax
	movl		N, %ebx

	movl		SEQ, %edi

outer_loop:

	/* Save outer loop counters */
	push		%eax
	push		%ebx

	/* Clear MMX registers */
	pxor		%mm1, %mm1
	pxor		%mm4, %mm4
	
	/* Calculate appropriate query_profile row */
	movzx		(%edi), %esi
	imul		M, %esi
	add			Q_PROFILE, %esi
	inc			%edi

	/* Reload FIH register */
	movl		FIH, %ecx

	/* for (j = 0; j < m / k; j++) */
	xor			%eax, %eax
	mov			M, %ebx
	shr			$3, %ebx

	/* mm0 - scorek
	   mm1 - ek
	   mm2 - fk
	   mm3 - hk
	   mm4 - x
	   mm5 - t
	   mm6 - gap_init
	   mm7 - gap_ext
	 */

inner_loop:

	/* hk <- H(j)
	   fk <- F(j) */
	movq		(%ecx), %mm2
	movq		SIMD_SIZE(%ecx), %mm3

	/* fk <- max { fk, hk - qk } - rk */
	movq		%mm3, %mm5
	psubusb		%mm6, %mm5
	PMAXUB		%mm5, %mm2
	psubusb		%mm7, %mm2

	/* t <- hk >> 7 * 8
	   hk <- x || hk << 1 * 8
	   x <- t */
	movq		%mm3, %mm5
	psrlq		$((SIMD_SIZE - 1) * 8), %mm5
	psllq		$8, %mm3
	por			%mm4, %mm3
	movq		%mm5, %mm4

	/* hk <- hk + s(c, j) - BIAS */
	paddusb		(%esi), %mm3
	psubusb		8 + SIMD_BIAS, %mm3

	/* hk <- max { hk, fk } */
	PMAXUB		%mm2, %mm3

	/* ek <- (hk << 1) || (ek >> 7) */
	movq		%mm3, %mm5
	psllq		$8, %mm5
	psrlq		$((SIMD_SIZE - 1) * 8), %mm1
	por			%mm5, %mm1

	/* ek <- ek - qk - rk */
	psubusb		%mm6, %mm1
	psubusb		%mm7, %mm1

	/* if (ek > 0) { 
	If condition removed: causes fewer loop mispredictions
	pxor		%mm5, %mm5
	psadbw		%mm1, %mm5
	pextrw		$0, %mm5, %edx
	cmp			$0, %edx

	ek == 0: 
	jle			zero
	*/

	/* ek > 0:
	   t <- ek */
	movq		%mm1, %mm5

/* while_loop:
	while (t > 0) {
		pxor		%mm8, %mm8
		psadbw		%mm5, %mm8
		pextrw		$0, %mm8, %edx
		cmp			$0, %edx

		jle			while_outside
	} jmp while_loop
	
	Unrolled loop follows
*/

	/* Iteration: 1 */
	/* t <- (t << 1) - rk */
	psllq		$8, %mm5
	psubusb		%mm7, %mm5
	/* ek <- max { ek, t} */
	PMAXUB		%mm5, %mm1

	/* Iteration: 2 */
	/* t <- (t << 1) - rk */
	psllq		$8, %mm5
	psubusb		%mm7, %mm5
	/* ek <- max { ek, t} */
	PMAXUB		%mm5, %mm1

	/* Iteration: 3 */
	/* t <- (t << 1) - rk */
	psllq		$8, %mm5
	psubusb		%mm7, %mm5
	/* ek <- max { ek, t} */
	PMAXUB		%mm5, %mm1

	/* Iteration: 4 */
	/* t <- (t << 1) - rk */
	psllq		$8, %mm5
	psubusb		%mm7, %mm5
	/* ek <- max { ek, t} */
	PMAXUB		%mm5, %mm1

	/* Iteration: 5 */
	/* t <- (t << 1) - rk */
	psllq		$8, %mm5
	psubusb		%mm7, %mm5
	/* ek <- max { ek, t} */
	PMAXUB		%mm5, %mm1

	/* Iteration: 6 */
	/* t <- (t << 1) - rk */
	psllq		$8, %mm5
	psubusb		%mm7, %mm5
	/* ek <- max { ek, t} */
	PMAXUB		%mm5, %mm1

	/* Iteration: 7 */
	/* t <- (t << 1) - rk */
	psllq		$8, %mm5
	psubusb		%mm7, %mm5
	/* ek <- max { ek, t} */
	PMAXUB		%mm5, %mm1

/* while_outside: */

	/* hk <- max { hk, ek } */
	PMAXUB		%mm1, %mm3

	/* ek <- max { hk, ek + qk } */
	paddusb		%mm6, %mm1
	PMAXUB		%mm3, %mm1

/*	If condition removed: causes fewer loop mispredictions
	jmp			outside

zero:
	t <= 0
	movq		%mm3, %mm1

outside:
*/

	/* H(j) <- hk
	   F(j) <- fk */
	movq		%mm2, (%ecx)
	movq		%mm3, SIMD_SIZE(%ecx)

	add			$(SIMD_SIZE * 2), %ecx
	add			$(SIMD_SIZE), %esi

	/ *scorek <- max { scorek, hk } */
	PMAXUB		%mm3, %mm0

	inc			%eax
	cmp			%ebx, %eax
	jl			inner_loop


	/* Restore outer loop counters */
	pop			%ebx
	pop			%eax

	inc			%eax
	cmp			%ebx, %eax
	jl			outer_loop

	/* Calculate final score */
	movq		%mm0, %mm1
	psrlq		$32, %mm0
	PMAXUB		%mm1, %mm0

	movq		%mm0, %mm1
	psrlq		$16, %mm0
	PMAXUB		%mm1, %mm0
	
	movq		%mm0, %mm1
	psrlq		$8, %mm0
	PMAXUB		%mm1, %mm0
	
	xor			%eax, %eax
	movd		%mm0, %ebx
	mov			%bl, %al
	movl		%eax, BIAS				/* Reuse stack variable */

#ifdef	DEBUG
	/* Done using MMX */
	emms
#endif

    /* Remove SIMD_BIAS */
	add			$(2 * 4), %esp

	pop			%eax			        /* Get alignment offset */
	add			%eax, %esp		    	/* Add offset */

	/* Function post processing */
	popa
	mov			BIAS, %eax
	leave
	ret
