
/*
 *  SIMD Smith-Waterman function for x86 platform
 *
 *  Technology: Intel SSE2
 *  Algorithm:  Diagonal
 *  Precision:  Byte
 *
 *  Copyright (C) 2003 Arpith Chacko Jacob
 *  Email: jarpith@acm.org
 *  Beta release 1.0.0.2
 *
 *  Start date:  25 june 2003
 *  Last Update: 08 september 2003
 *
 */

#define QUERY		8(%ebp)
#define M			12(%ebp)
#define SEQ			16(%ebp)
#define N			20(%ebp)
#define Q_PROFILE	24(%ebp)
#define BIAS		28(%ebp)
#define GAP_INIT	32(%ebp)
#define GAP_EXT		36(%ebp)
#define FIH			40(%ebp)

#define SIMD_BIAS	4 * SIMD_SIZE(%esp)
#define SIMD_QPP_MEM	0(%esp)

#define SIMD_SIZE	16

.global _sw_simd_sse2_diag_b, sw_simd_sse2_diag_b


.text

/* 
 * function _sw_simd_sse2_diag_b
 * C prototype
 * int sw_simd_sse2_diag_b(char *query, int m, char *seq, int n, char *query_profile,
 *                         int bias, int gap_init, int gap_ext, char *fih);
 *
 */

sw_simd_sse2_diag_b:
_sw_simd_sse2_diag_b:

	/* Function preprocessing */
	enter		$0, $0
	pusha

	/* Calculate the adjusted (with dummy chars) length of the query sequence */
	add			$(SIMD_SIZE * 2 - 2), M

	/* Align stack pointer for SIMD_SIZE (2^x) operations */
	movl		%esp, %eax
	sub			$4, %eax			    /* Size of push eax */
	and			$(SIMD_SIZE - 1), %eax	/* Calculate alignment offset */
	sub			%eax, %esp		    	/* Add offset */
	push		%eax			        /* Store offset */

	/* Load bias */
	movb		BIAS, %al
	movb		%al, %ah
	shl			$16, %eax
	movb		BIAS, %al
	movb		%al, %ah
	push		%eax                 	/* Store BIAS in SIMD_SIZE aligned */
	push		%eax					/* location 					   */
	push		%eax                 	/* Totally 128 bits                */
	push		%eax

	/* Space for SIMD_SM_MEM */
	sub			$(SIMD_SIZE * 4), %esp

	/* Load gap_init, gap_ext */
	pxor		%xmm6, %xmm6
	pxor		%xmm7, %xmm7

	movb		GAP_INIT, %al
	movb		%al, %ah
    pinsrw      $0, %eax, %xmm6
    pinsrw      $1, %eax, %xmm6
    pinsrw      $2, %eax, %xmm6
    pinsrw      $3, %eax, %xmm6
    pinsrw      $4, %eax, %xmm6
    pinsrw      $5, %eax, %xmm6
    pinsrw      $6, %eax, %xmm6
    pinsrw      $7, %eax, %xmm6

	movb		GAP_EXT, %al
	movb		%al, %ah
    pinsrw      $0, %eax, %xmm7
    pinsrw      $1, %eax, %xmm7
    pinsrw      $2, %eax, %xmm7
    pinsrw      $3, %eax, %xmm7
    pinsrw      $4, %eax, %xmm7
    pinsrw      $5, %eax, %xmm7
    pinsrw      $6, %eax, %xmm7
    pinsrw      $7, %eax, %xmm7

	/* for (i = 0; i < m + 2 * SIMD_SIZE - 2; i++) {
		 F(i) = 0;
	 	 H(i) = 0;
	   }
	*/

	xor			%eax, %eax
	movl		M, %ebx

	movl		FIH, %ecx

init_loop:
	movw		$0, (%ecx, %eax, 2)

	inc			%eax
	cmp			%ebx, %eax
	jl			init_loop

	/* score = 0 */
	pxor		%xmm0, %xmm0

	/* for (i = 0; i < n/16; i++) */
	xor			%eax, %eax
	movl		N, %ebx
	shr			$4, %ebx			    /* SIMD = 16, shift right */

	movl		SEQ, %edi

outer_loop:

	/* Save outer loop counters */
	push		%eax
	push		%ebx

	/* Clear XMM registers */
	pxor		%xmm1, %xmm1
	pxor		%xmm4, %xmm4
	pxor		%xmm5, %xmm5
	
	/* fk <- [ F(k - 1), 0, 0, 0, 0, 0, 0, 0 ] */
	pxor		%xmm2, %xmm2
	xor			%eax, %eax
	movb		((SIMD_SIZE - 1) * 2)(%ecx), %ah
	pinsrw		$7, %eax, %xmm2

	/* hk <- [ H(k - 1), 0, 0, 0, 0, 0, 0, 0 ] */
	pxor		%xmm3, %xmm3
	movb		((SIMD_SIZE - 1) * 2 + 1)(%ecx), %ah
	pinsrw		$7, %eax, %xmm3

	/* Calculate query_profile scores for next k set */
	movl		Q_PROFILE, %eax

	movzx		0(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 0), %edx
	add			%eax, %edx
	movl		%edx, 8 + SIMD_QPP_MEM

	movzx		1(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 1), %edx
	add			%eax, %edx
	movl		%edx, 4 + 8 + SIMD_QPP_MEM

	movzx		2(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 2), %edx
	add			%eax, %edx
	movl		%edx, 8 + 8 + SIMD_QPP_MEM

	movzx		3(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 3), %edx
	add			%eax, %edx
	movl		%edx, 12 + 8 + SIMD_QPP_MEM

	movzx		4(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 4), %edx
	add			%eax, %edx
	movl		%edx, 16 + 8 + SIMD_QPP_MEM

	movzx		5(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 5), %edx
	add			%eax, %edx
	movl		%edx, 20 + 8 + SIMD_QPP_MEM

	movzx		6(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 6), %edx
	add			%eax, %edx
	movl		%edx, 24 + 8 + SIMD_QPP_MEM

	movzx		7(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 7), %edx
	add			%eax, %edx
	movl		%edx, 28 + 8 + SIMD_QPP_MEM

	movzx		8(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 8), %edx
	add			%eax, %edx
	movl		%edx, 32 + 8 + SIMD_QPP_MEM

	movzx		9(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 9), %edx
	add			%eax, %edx
	movl		%edx, 36 + 8 + SIMD_QPP_MEM

	movzx		10(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 10), %edx
	add			%eax, %edx
	movl		%edx, 40 + 8 + SIMD_QPP_MEM

	movzx		11(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 11), %edx
	add			%eax, %edx
	movl		%edx, 44 + 8 + SIMD_QPP_MEM

	movzx		12(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 12), %edx
	add			%eax, %edx
	movl		%edx, 48 + 8 + SIMD_QPP_MEM

	movzx		13(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 13), %edx
	add			%eax, %edx
	movl		%edx, 52 + 8 + SIMD_QPP_MEM

	movzx		14(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 14), %edx
	add			%eax, %edx
	movl		%edx, 56 + 8 + SIMD_QPP_MEM

	movzx		15(%edi), %edx
	imul		M, %edx
//	add			$(SIMD_SIZE - 1 - 15), %edx
//						^^^ ---> evaluates to zero
	add			%eax, %edx
	movl		%edx, 60 + 8 + SIMD_QPP_MEM

    xor			%esi, %esi

	/* for (j = 0; j < m+k-1; j++) */
	xor			%eax, %eax
	movl		M, %ebx
	sub			$(SIMD_SIZE - 1), %ebx

	/* xmm0 - scorek
	   xmm1 - ek
	   xmm2 - fk
	   xmm3 - hk
	   xmm4 - hck
	   xmm5 - hdk
	   xmm6 - gap_init
	   xmm7 - gap_ext
	 */

inner_loop:
	/* Push loop counters */
	push		%ebx
	push		%eax

	/* ek <- max { ek, hck - qk } - rk */
	psubusb		%xmm6, %xmm4
	pmaxub		%xmm4, %xmm1
	psubusb		%xmm7, %xmm1

	/* fk <- max { fk, hk - qk } - rk */
	movdqa		%xmm3, %xmm4
	psubusb		%xmm6, %xmm4
	pmaxub		%xmm4, %xmm2
	psubusb		%xmm7, %xmm2

	/* hck <- max { 0, ek, hdk + sim_mat(), fk }
	   Sim matrix in mm4
	   s(k * i + 0, j + k - 0 )
	   s(k * i + 1, j + k - 1 )
	   ..
	   ..
	   s(k * i + 7, j + k - 7 )
	 */

	movl		16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %ah

	movl		4 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %al

	movl		8 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bh

	movl		12 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bl

	pinsrw		$7, %eax, %xmm4
	pinsrw		$6, %ebx, %xmm4

	movl		16 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %ah

	movl		20 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %al

	movl		24 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bh

	movl		28 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bl

	pinsrw		$5, %eax, %xmm4
	pinsrw		$4, %ebx, %xmm4

	movl		32 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %ah

	movl		36 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %al

	movl		40 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bh

	movl		44 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bl

	pinsrw		$3, %eax, %xmm4
	pinsrw		$2, %ebx, %xmm4

	movl		48 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %ah

	movl		52 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %al

	movl		56 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bh

	movl		60 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bl

	pinsrw		$1, %eax, %xmm4
	pinsrw		$0, %ebx, %xmm4

	/* Increment pointer to next position in query profile */
	inc			%esi

	/* hck <- max { 0, ek, hdk + sim_mat(), fk }
	   hdk + sim_mat()
	 */
	paddusb		%xmm5, %xmm4

	/* hdk + sim_mat() - BIAS */
	psubusb		16 + SIMD_BIAS, %xmm4

	/* hck <- max { hck, ek, fk } */
	pmaxub		%xmm1, %xmm4
	pmaxub		%xmm2, %xmm4

	/* scorek <- max { scorek, hck } */
	pmaxub		%xmm4, %xmm0

	/* hdk <- hk */
	movdqa		%xmm3, %xmm5

	/* Pop loop counter for use below */
	pop			%eax

	/* F(j) <- fk(8)
	   fk <- F(j + k) | fk(1..7)
	 */
	pextrw		$0, %xmm2, %edx
	psrldq		$1, %xmm2
	movb		%dl, (%ecx, %eax, 2)
	pextrw		$7, %xmm2, %edx
	movb		(SIMD_SIZE * 2)(%ecx, %eax, 2), %dh

	/* H(j) <- hck(8)
	   hk <- H(j + k) | hck(1..7)
	 */
	movdqa		%xmm4, %xmm3
	pextrw		$0, %xmm3, %ebx
	psrldq		$1, %xmm3
	movb		%bl, 1(%ecx, %eax, 2)
	pextrw		$7, %xmm3, %ebx
	movb		(SIMD_SIZE * 2 + 1)(%ecx, %eax, 2), %bh

	pinsrw		$7, %edx, %xmm2
	pinsrw		$7, %ebx, %xmm3

	/* Pop loop limit */
	pop			%ebx

	inc			%eax
	cmp			%ebx, %eax
	jl			inner_loop


	/* Restore outer loop counters */
	pop			%ebx
	pop			%eax

	/* Increment seq counter to next k set */
	add			$(SIMD_SIZE), %edi

	inc			%eax
	cmp			%ebx, %eax
	jl			outer_loop


	/* Calculate final score */
	movdqa		%xmm0, %xmm1
	psrldq		$8, %xmm0
	pmaxub		%xmm1, %xmm0

	movq		%xmm0, %xmm1
	psrldq		$4, %xmm0
	pmaxub		%xmm1, %xmm0

	movq		%xmm0, %xmm1
	psrldq		$2, %xmm0
	pmaxub		%xmm1, %xmm0
	
	movq		%xmm0, %xmm1
	psrldq		$1, %xmm0
	pmaxub		%xmm1, %xmm0

	xor			%eax, %eax
	movd		%xmm0, %ebx
	mov			%bl, %al
	movl		%eax, BIAS				/* Reuse stack variable */

    /* Remove SIMD_BIAS/SIMD_SM_MEM */
	add			$((SIMD_SIZE + 4) * 4), %esp

	pop			%eax			        /* Get alignment offset */
	add			%eax, %esp		    	/* Add offset */

	/* Function post processing */
	popa
	mov			BIAS, %eax
	leave
	ret
