
/*
 *  SIMD Smith-Waterman function for x86 platform
 *
 *  Technology: Intel SSE2
 *  Algorithm:  Horizontal
 *  Precision:  Byte
 *
 *  Copyright (C) 2003 Arpith Chacko Jacob
 *  Email: jarpith@acm.org
 *  Beta release 1.0.0.2
 *
 *  Start date:  29 june 2003
 *  Last Update: 08 september 2003
 *
 */

#define QUERY		8(%ebp)
#define M			12(%ebp)
#define SEQ			16(%ebp)
#define N			20(%ebp)
#define Q_PROFILE	24(%ebp)
#define BIAS		28(%ebp)
#define GAP_INIT	32(%ebp)
#define GAP_EXT		36(%ebp)
#define FIH			40(%ebp)

#define SIMD_BIAS	0(%esp)

#define SIMD_SIZE	16

.global _sw_simd_sse2_horiz_b, sw_simd_sse2_horiz_b


.text

/* 
 * function _sw_simd_sse2_horiz_b
 * C prototype
 * int sw_simd_sse2_horiz_b(char *query, int m, char *seq, int n, char *query_profile,
 *                          int bias, int gap_init, int gap_ext, char *fih);
 *
 */

sw_simd_sse2_horiz_b:
_sw_simd_sse2_horiz_b:

	/* Function preprocessing */
	enter		$0, $0
	pusha

	/* Align stack pointer for SIMD_SIZE (2^x) operations */
	movl		%esp, %eax
	sub			$4, %eax			    /* Size of push eax */
	and			$(SIMD_SIZE - 1), %eax	/* Calculate alignment offset */
	sub			%eax, %esp		    	/* Subtract offset */
	push		%eax			        /* Store offset */

	/* Load bias */
	movb		BIAS, %al
	movb		%al, %ah
	shl			$16, %eax
	movb		BIAS, %al
	movb		%al, %ah
	push		%eax                 	/* Store BIAS in SIMD_SIZE aligned */
	push		%eax					/* location 					   */
	push		%eax                 	/* Totally 128 bits                */
	push		%eax

	/* Load gap_init, gap_ext */
	pxor		%xmm6, %xmm6
	pxor		%xmm7, %xmm7

	movb		GAP_INIT, %al
	movb		%al, %ah
    pinsrw      $0, %eax, %xmm6
    pinsrw      $1, %eax, %xmm6
    pinsrw      $2, %eax, %xmm6
    pinsrw      $3, %eax, %xmm6
    pinsrw      $4, %eax, %xmm6
    pinsrw      $5, %eax, %xmm6
    pinsrw      $6, %eax, %xmm6
    pinsrw      $7, %eax, %xmm6

	movb		GAP_EXT, %al
	movb		%al, %ah
    pinsrw      $0, %eax, %xmm7
    pinsrw      $1, %eax, %xmm7
    pinsrw      $2, %eax, %xmm7
    pinsrw      $3, %eax, %xmm7
    pinsrw      $4, %eax, %xmm7
    pinsrw      $5, %eax, %xmm7
    pinsrw      $6, %eax, %xmm7
    pinsrw      $7, %eax, %xmm7

	/* for (i = 0; i < m; i++) {
		 F(i) = 0;
	 	 H(i) = 0;
	   }
	*/

	xor			%eax, %eax
	movl		M, %ebx

	movl		FIH, %ecx

init_loop:
	movw		$0, (%ecx, %eax, 2)

	inc			%eax
	cmp			%ebx, %eax
	jl			init_loop

	/* score = 0 */
	pxor		%xmm0, %xmm0

	/* for (i = 0; i < n; i++) */
	xor			%eax, %eax
	movl		N, %ebx

	movl		SEQ, %edi

outer_loop:

	/* Save outer loop counters */
	push		%eax
	push		%ebx

	/* Clear MMX registers */
	pxor		%xmm1, %xmm1
	pxor		%xmm4, %xmm4
	
	/* Calculate appropriate query_profile row */
	movzx		(%edi), %esi
	imul		M, %esi
	add			Q_PROFILE, %esi
	inc			%edi

	/* Reload FIH register */
	movl		FIH, %ecx

	/* for (j = 0; j < m / k; j++) */
	xor			%eax, %eax
	mov			M, %ebx
	shr			$4, %ebx

	/* xmm0 - scorek
	   xmm1 - ek
	   xmm2 - fk
	   xmm3 - hk
	   xmm4 - x
	   xmm5 - t
	   xmm6 - gap_init
	   xmm7 - gap_ext
	 */

inner_loop:

	/* hk <- H(j)
	   fk <- F(j) */
	movdqa		(%ecx), %xmm2
	movdqa		SIMD_SIZE(%ecx), %xmm3

	/* fk <- max { fk, hk - qk } - rk */
	movdqa		%xmm3, %xmm5
	psubusb		%xmm6, %xmm5
	pmaxub		%xmm5, %xmm2
	psubusb		%xmm7, %xmm2

	/* t <- hk >> 7 * 8
	   hk <- x || hk << 1 * 8
	   x <- t */
	movdqa		%xmm3, %xmm5
	psrldq		$(SIMD_SIZE - 1), %xmm5
	pslldq		$1, %xmm3
	por			%xmm4, %xmm3
	movdqa		%xmm5, %xmm4

	/* hk <- hk + s(c, j) - BIAS */
	paddusb		(%esi), %xmm3
	psubusb		8 + SIMD_BIAS, %xmm3

	/* hk <- max { hk, fk } */
	pmaxub		%xmm2, %xmm3

	/* ek <- (hk << 1) || (ek >> 7) */
	movdqa		%xmm3, %xmm5
	pslldq		$1, %xmm5
	psrldq		$(SIMD_SIZE - 1), %xmm1
	por			%xmm5, %xmm1

	/* ek <- ek - qk - rk */
	psubusb		%xmm6, %xmm1
	psubusb		%xmm7, %xmm1

	/* if (ek > 0) { */
	pxor		%xmm5, %xmm5
	psadbw		%xmm1, %xmm5
	pextrw		$0, %xmm5, %edx
	cmp			$0, %edx
	jg			not_zero
	pextrw		$4, %xmm5, %edx
	cmp			$0, %edx

	/* ek == 0: */
	je			zero

not_zero:
	/* ek > 0:
	   t <- ek */
	movdqa		%xmm1, %xmm5

/* while_loop:
	while (t > 0) {
		pxor		%xmm8, %xmm8
		psadbw		%xmm5, %xmm8
		pextrw		$0, %xmm8, %edx
		cmp			$0, %edx

		jle			while_outside
	} jmp while_loop
	
	Unrolled loop follows
*/

	/* Iteration: 1 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 2 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 3 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 4 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 5 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 6 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 7 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 8 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 9 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 10 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 11 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 12 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 13 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 14 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

	/* Iteration: 15 */
	/* t <- (t << 1) - rk */
	pslldq		$1, %xmm5
	psubusb		%xmm7, %xmm5
	/* ek <- max { ek, t} */
	pmaxub		%xmm5, %xmm1

/* while_outside: */

	/* hk <- max { hk, ek } */
	pmaxub		%xmm1, %xmm3

	/* ek <- max { hk, ek + qk } */
	paddusb		%xmm6, %xmm1
	pmaxub		%xmm3, %xmm1

	jmp			outside

zero:
	/* t <= 0 */
	movdqa		%xmm3, %xmm1

outside:

	/* H(j) <- hk
	   F(j) <- fk */
	movdqa		%xmm2, (%ecx)
	movdqa		%xmm3, SIMD_SIZE(%ecx)

	add			$(SIMD_SIZE * 2), %ecx
	add			$(SIMD_SIZE), %esi

	/ *scorek <- max { scorek, hk } */
	pmaxub		%xmm3, %xmm0

	inc			%eax
	cmp			%ebx, %eax
	jl			inner_loop


	/* Restore outer loop counters */
	pop			%ebx
	pop			%eax

	inc			%eax
	cmp			%ebx, %eax
	jl			outer_loop

	/* Calculate final score */
	movdqa		%xmm0, %xmm1
	psrldq		$8, %xmm0
	pmaxub		%xmm1, %xmm0

	movq		%xmm0, %xmm1
	psrldq		$4, %xmm0
	pmaxub		%xmm1, %xmm0

	movq		%xmm0, %xmm1
	psrldq		$2, %xmm0
	pmaxub		%xmm1, %xmm0
	
	movq		%xmm0, %xmm1
	psrldq		$1, %xmm0
	pmaxub		%xmm1, %xmm0

	xor			%eax, %eax
	movd		%xmm0, %ebx
	mov			%bl, %al
	movl		%eax, BIAS				/* Reuse stack variable */

    /* Remove SIMD_BIAS */
	add			$(4 * 4), %esp

	pop			%eax			        /* Get alignment offset */
	add			%eax, %esp		    	/* Add offset */

	/* Function post processing */
	popa
	mov			BIAS, %eax
	leave
	ret
