
/*
 *  SIMD Smith-Waterman function for x86 platform
 *
 *  Technology: Intel SSE
 *  Algorithm:  Diagonal
 *  Precision:  Byte
 *
 *  Copyright (C) 2003 Arpith Chacko Jacob
 *  Email: jarpith@acm.org
 *  Beta release 1.0.0.2
 *
 *  Start date:  20 june 2003
 *  Last Update: 08 september 2003
 *
 */

#define QUERY		8(%ebp)
#define M			12(%ebp)
#define SEQ			16(%ebp)
#define N			20(%ebp)
#define Q_PROFILE	24(%ebp)
#define BIAS		28(%ebp)
#define GAP_INIT	32(%ebp)
#define GAP_EXT		36(%ebp)
#define FIH			40(%ebp)

#define SIMD_BIAS	4 * SIMD_SIZE(%esp)
#define SIMD_QPP_MEM	0(%esp)

#define SIMD_SIZE	8

.global _sw_simd_sse_diag_b, sw_simd_sse_diag_b


.text

/* 
 * function _sw_simd_sse_diag_b
 * C prototype
 * int sw_simd_sse_diag_b(char *query, int m, char *seq, int n, char *query_profile,
 *                        int bias, int gap_init, int gap_ext, char *fih);
 *
 */

sw_simd_sse_diag_b:
_sw_simd_sse_diag_b:

	/* Function preprocessing */
	enter		$0, $0
	pusha

	/* Calculate the adjusted (with dummy chars) length of the query sequence */
	add			$(SIMD_SIZE * 2 - 2), M

	/* Align stack pointer for SIMD_SIZE (2^x) operations */
	movl		%esp, %eax
	sub			$4, %eax			    /* Size of push eax */
	and			$(SIMD_SIZE - 1), %eax	/* Calculate alignment offset */
	sub			%eax, %esp		    	/* Subtract offset */
	push		%eax			        /* Store offset */

	/* Load bias */
	movb		BIAS, %al
	movb		%al, %ah
	shl			$16, %eax
	movb		BIAS, %al
	movb		%al, %ah
	push		%eax                 	/* Store BIAS in SIMD_SIZE aligned */
	push		%eax					/* location 					   */

	/* Space for SIMD_SM_MEM */
	sub			$(SIMD_SIZE * 4), %esp

	/* Load gap_init, gap_ext */
	pxor		%mm6, %mm6
	pxor		%mm7, %mm7

	movb		GAP_INIT, %al
	movb		%al, 1 + GAP_INIT
	pshufw		$0, GAP_INIT, %mm6

	movb		GAP_EXT, %al
	movb		%al, 1 + GAP_EXT
	pshufw		$0, GAP_EXT, %mm7

	/* for (i = 0; i < m + 2 * SIMD_SIZE - 2; i++) {
		 F(i) = 0;
	 	 H(i) = 0;
	   }
	*/

	xor			%eax, %eax
	movl		M, %ebx

	movl		FIH, %ecx

init_loop:
	movw		$0, (%ecx, %eax, 2)

	inc			%eax
	cmp			%ebx, %eax
	jl			init_loop

	/* score = 0 */
	pxor		%mm0, %mm0

	/* for (i = 0; i < n/8; i++) */
	xor			%eax, %eax
	movl		N, %ebx
	shr			$3, %ebx			    /* SIMD = 8, shift right thrice */

	movl		SEQ, %edi

outer_loop:

	/* Save outer loop counters */
	push		%eax
	push		%ebx

	/* Clear MMX registers */
	pxor		%mm1, %mm1
	pxor		%mm4, %mm4
	pxor		%mm5, %mm5
	
	/* fk <- [ F(k - 1), 0, 0, 0, 0, 0, 0, 0 ] */
	movb		((SIMD_SIZE - 1) * 2)(%ecx), %al
	movd		%eax, %mm2
	psllq		$56, %mm2

	/* hk <- [ H(k - 1), 0, 0, 0, 0, 0, 0, 0 ] */
	movb		((SIMD_SIZE - 1) * 2 + 1)(%ecx), %al
	movd		%eax, %mm3
	psllq		$56, %mm3

	/* Calculate query_profile scores for next k set */
	movl		Q_PROFILE, %eax

	movzx		0(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 0), %edx
	add			%eax, %edx
	movl		%edx, 8 + SIMD_QPP_MEM

	movzx		1(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 1), %edx
	add			%eax, %edx
	movl		%edx, 4 + 8 + SIMD_QPP_MEM

	movzx		2(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 2), %edx
	add			%eax, %edx
	movl		%edx, 8 + 8 + SIMD_QPP_MEM

	movzx		3(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 3), %edx
	add			%eax, %edx
	movl		%edx, 12 + 8 + SIMD_QPP_MEM

	movzx		4(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 4), %edx
	add			%eax, %edx
	movl		%edx, 16 + 8 + SIMD_QPP_MEM

	movzx		5(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 5), %edx
	add			%eax, %edx
	movl		%edx, 20 + 8 + SIMD_QPP_MEM

	movzx		6(%edi), %edx
	imul		M, %edx
	add			$(SIMD_SIZE - 1 - 6), %edx
	add			%eax, %edx
	movl		%edx, 24 + 8 + SIMD_QPP_MEM

	movzx		7(%edi), %edx
	imul		M, %edx
//	add			$(SIMD_SIZE - 1 - 7), %edx
//						^^^ ---> evaluates to zero
	add			%eax, %edx
	movl		%edx, 28 + 8 + SIMD_QPP_MEM

    xor			%esi, %esi

	/* for (j = 0; j < m+k-1; j++) */
	xor			%eax, %eax
	movl		M, %ebx
	sub			$(SIMD_SIZE - 1), %ebx

	/* mm0 - scorek
	   mm1 - ek
	   mm2 - fk
	   mm3 - hk
	   mm4 - hck
	   mm5 - hdk
	   mm6 - gap_init
	   mm7 - gap_ext
	 */

inner_loop:
	/* Push loop counters */
	push		%ebx
	push		%eax

	/* ek <- max { ek, hck - qk } - rk */
	psubusb		%mm6, %mm4
	pmaxub		%mm4, %mm1
	psubusb		%mm7, %mm1

	/* fk <- max { fk, hk - qk } - rk */
	movq		%mm3, %mm4
	psubusb		%mm6, %mm4
	pmaxub		%mm4, %mm2
	psubusb		%mm7, %mm2

	/* hck <- max { 0, ek, hdk + sim_mat(), fk }
	   Sim matrix in mm4
	   s(k * i + 0, j + k - 0 )
	   s(k * i + 1, j + k - 1 )
	   ..
	   ..
	   s(k * i + 7, j + k - 7 )
	 */

	movl		16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %ah

	movl		4 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %al

	movl		8 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bh

	movl		12 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bl

	pinsrw		$3, %eax, %mm4
	pinsrw		$2, %ebx, %mm4

	movl		16 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %ah

	movl		20 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %al

	movl		24 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bh

	movl		28 + 16 + SIMD_QPP_MEM, %edx
	movb		(%edx, %esi, 1), %bl

	pinsrw		$1, %eax, %mm4
	pinsrw		$0, %ebx, %mm4

	/* Increment pointer to next position in query profile */
	inc			%esi

	/* hck <- max { 0, ek, hdk + sim_mat(), fk }
	   hdk + sim_mat()
	 */
	paddusb		%mm5, %mm4

	/* hdk + sim_mat() - BIAS */
	psubusb		16 + SIMD_BIAS, %mm4

	/* hck <- max { hck, ek, fk } */
	pmaxub		%mm1, %mm4
	pmaxub		%mm2, %mm4

	/* scorek <- max { scorek, hck } */
	pmaxub		%mm4, %mm0

	/* hdk <- hk */
	movq		%mm3, %mm5

	/* Pop loop counter for use below */
	pop			%eax

	/* F(j) <- fk(8)
	   fk <- F(j + k) | fk(1..7)
	 */
	pextrw		$0, %mm2, %edx
	psrlq		$8, %mm2
	movb		%dl, (%ecx, %eax, 2)
	pextrw		$3, %mm2, %edx
	movb		(SIMD_SIZE * 2)(%ecx, %eax, 2), %dh

	/* H(j) <- hck(8)
	   hk <- H(j + k) | hck(1..7)
	 */
	movq		%mm4, %mm3
	pextrw		$0, %mm3, %ebx
	psrlq		$8, %mm3
	movb		%bl, 1(%ecx, %eax, 2)
	pextrw		$3, %mm3, %ebx
	movb		(SIMD_SIZE * 2 + 1)(%ecx, %eax, 2), %bh

	pinsrw		$3, %edx, %mm2
	pinsrw		$3, %ebx, %mm3

	/* Pop loop limit */
	pop			%ebx

	inc			%eax
	cmp			%ebx, %eax
	jl			inner_loop


	/* Restore outer loop counters */
	pop			%ebx
	pop			%eax

	/* Increment seq counter to next k set */
	add			$(SIMD_SIZE), %edi

	inc			%eax
	cmp			%ebx, %eax
	jl			outer_loop

	/* Calculate final score */
	movq		%mm0, %mm1
	psrlq		$32, %mm0
	pmaxub		%mm1, %mm0

	movq		%mm0, %mm1
	psrlq		$16, %mm0
	pmaxub		%mm1, %mm0
	
	movq		%mm0, %mm1
	psrlq		$8, %mm0
	pmaxub		%mm1, %mm0
	
	xor			%eax, %eax
	movd		%mm0, %ebx
	mov			%bl, %al
	movl		%eax, BIAS				/* Reuse stack variable */

#ifdef	DEBUG
	/* Done using MMX */
	emms
#endif

    /* Remove SIMD_BIAS/SIMD_SM_MEM */
	add			$((SIMD_SIZE + 2) * 4), %esp

	pop			%eax			        /* Get alignment offset */
	add			%eax, %esp		    	/* Add offset */

	/* Function post processing */
	popa
	mov			BIAS, %eax
	leave
	ret
