
/*
 *  SIMD Smith-Waterman functions for x86 platform
 *
 *  Technology: Intel MMX, SSE, SSE2
 *  Algorithms: Horizontal (optimum for large affine gap penalties)
 *              Diagonal (consistent speed for all gap penalties)
 *  Precision:  Byte, Word
 *
 *  Copyright (C) 2003 Arpith Chacko Jacob
 *  Email: jarpith@acm.org
 *  Beta release 1.0.0.2
 *
 *  Start date:  20 june 2003
 *  Last Update: 08 september 2003
 *
 */

#ifndef __SW_SIMD_X86_H__
#define __SW_SIMD_X86_H__

/* SIMD technology detected */
#define	SW_SIMD_NO_SIMD			0
#define	SW_SIMD_MMX				1
#define	SW_SIMD_SSE				2
#define	SW_SIMD_SSE2			3

/* S-W score precision required */
#define	SW_SIMD_BYTE			1
#define	SW_SIMD_WORD			2

/* Vector lengths of SIMD words */
#define	SW_SIMD_MMX_BYTE_VLEN	8
#define	SW_SIMD_MMX_WORD_VLEN	4
#define	SW_SIMD_SSE_BYTE_VLEN	8
#define	SW_SIMD_SSE_WORD_VLEN	4
#define	SW_SIMD_SSE2_BYTE_VLEN	16
#define	SW_SIMD_SSE2_WORD_VLEN	8

/* Saturation value of byte precision */
#define	SW_SIMD_BYTE_SATURATION	255

/* Search type */
#define	SW_DNA_SEARCH			0
#define	SW_PROTEIN_SEARCH		1

#define ID_LEN					20
#define BUF_INCR				10240
#define QBUF_LEN				102400
#define SBUF_LEN				102400

/* Function to detect best SIMD technology */
inline int get_simd_tech ();

/* Timing measurement functions */
inline void start_counter (unsigned int *, unsigned int *);
inline double end_counter (unsigned int *, unsigned int *,
			  			   unsigned int *, unsigned int *);

/* Sequence support functions */
void *memalign (char *[2], int, unsigned int);
inline int next_seq (FILE *, int, char *, char *, int);
inline char DNA2int (char);
inline char AA2int (char);

#endif	/* ifndef __SW_SIMD_X86_H__ */
